/*
 * GQradio
 * (C) 2004 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */


#include "gqradio.h"
#include "preferences.h"

#include "display.h"
#include "io_radio.h"
#include "mixer.h"
#include "preset.h"
#include "ui2_editor.h"
#include "ui2_main.h"
#include "ui_fileops.h"
#include "ui_menu.h"
#include "ui_tabcomp.h"
#include "ui_utildlg.h"
#include "ui_tree_edit.h"


#define PREF_WINDOW_DEFAULT_WIDTH  350
#define PREF_WINDOW_DEFAULT_HEIGHT 450


enum {
	SKIN_COLUMN_PATH = 0,
	SKIN_COLUMN_NAME,
	SKIN_COLUMN_COUNT
};

enum {
	PRESET_COLUMN_NUMBER = 0,
	PRESET_COLUMN_FREQUENCY,
	PRESET_COLUMN_DESCRIPTION,
	PRESET_COLUMN_PATH,
	PRESET_COLUMN_COUNT
};


static GtkWidget *configwindow;
static GtkWidget *mixer_entry;
static GtkWidget *default_skin_entry;
static GtkWidget *preset_view;
static GtkWidget *device_entry;

static gint c_slik_smart_placement;
static gint c_slik_remember_position;
static gint c_mixer_device_id;
static gint c_wm_decorations;
static gint c_slik_double_size;
static gint c_description_scrolls_always;
static gint c_skinned_menus_enable;

static gint c_slik_colorshift_on;
static gint c_slik_colorshift_r;
static gint c_slik_colorshift_g;
static gint c_slik_colorshift_b;
static gint c_slik_colorshift_a;

static gint c_slik_transparency_force;
static gint c_slik_transparency_force_a;

static gint c_use_reported_freq_limits;

static gint c_freq_step;


static void create_config_window(gint start_tab);


static void hscale_centered_cb(GtkObject *adj, gpointer data)
{
	gint *c_n = data;

	*c_n = (gint)(GTK_ADJUSTMENT(adj)->value / 200.0 * 255.0 + 128.0);
}

static void hscale_cb(GtkObject *adj, gpointer data)
{
	gint *c_n = data;

	*c_n = (gint)(GTK_ADJUSTMENT(adj)->value / 100.0 * 255.0);
}

static GtkWidget *hscale_new(GtkWidget *box, const gchar *text, gint n, gint *c_n, gint centered)
{
	GtkObject *adj;
        GtkWidget *hscale;

	*c_n = n;

	if (centered)
		{
		hscale = ui_edit_spin_new_with_scale(box, text, -100, 100, &adj);
		gtk_adjustment_set_value(GTK_ADJUSTMENT(adj), ((float)n - 128.0) / 255.0 * 200.0);
		}
	else
		{
		hscale = ui_edit_spin_new_with_scale(box, text, 0, 100, &adj);
		gtk_adjustment_set_value(GTK_ADJUSTMENT(adj), (float)n / 255.0 * 100.0);
		}

	if (centered)
		{
	        g_signal_connect(G_OBJECT(adj), "value_changed",
				 G_CALLBACK(hscale_centered_cb), c_n);
		}
	else
		{
		g_signal_connect(G_OBJECT(adj), "value_changed",
				 G_CALLBACK(hscale_cb), c_n);
		}

	return hscale;
}

static void checkbox_cb(GtkWidget *w, gpointer data)
{
	gint *c_n = data;

	*c_n = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
}

static GtkWidget *checkbox_new(gchar *text, gint n, gint *c_n)
{
        GtkWidget *button;

	*c_n = n;
	button = gtk_check_button_new_with_label(text);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), n);
	g_signal_connect(G_OBJECT(button), "clicked",
			 G_CALLBACK(checkbox_cb), c_n);

	return button;
}

static gint preset_view_frequency_cb(TreeEditData *ted, const gchar *oldname, const gchar *newname, gpointer data)
{
	GtkTreeModel *store;
	GtkTreeIter iter;
	guint32 freq;
	gchar *old_path;

	freq = radio_freq_from_text(newname);

	store = gtk_tree_view_get_model(ted->tree);
	gtk_tree_model_get_iter(store, &iter, ted->path);
	gtk_tree_model_get(store, &iter, PRESET_COLUMN_PATH, &old_path, -1);
	if (freq < radio_limit_get_lower() || freq > radio_limit_get_upper())
		{
		gtk_list_store_set(GTK_LIST_STORE(store), &iter,
				   PRESET_COLUMN_FREQUENCY, NULL,
				   PRESET_COLUMN_DESCRIPTION, NULL,
				   PRESET_COLUMN_PATH, "clear", -1);
		}
	else
		{
		gchar *new_path;
		gchar *freq_text;
		gchar *tmp;

		if (old_path)
			{
			tmp = radio_path_to_title(old_path);
			}
		else
			{
			tmp = g_strdup(preset_get_description(tree_path_to_row(ted->path)));
			}
		new_path = radio_freq_to_path(freq, tmp);
		g_free(tmp);

		freq_text = radio_freq_to_text(freq);
		gtk_list_store_set(GTK_LIST_STORE(store), &iter,
				   PRESET_COLUMN_FREQUENCY, freq_text,
				   PRESET_COLUMN_PATH, new_path, -1);
		g_free(freq_text);
		g_free(new_path);
		}

	g_free(old_path);

	return FALSE;
}

static gint preset_view_description_cb(TreeEditData *ted, const gchar *oldname, const gchar *newname, gpointer data)
{
	GtkTreeModel *store;
	GtkTreeIter iter;
	gchar *old_path;
	gchar *new_path;

	store = gtk_tree_view_get_model(ted->tree);
	gtk_tree_model_get_iter(store, &iter, ted->path);
	gtk_tree_model_get(store, &iter, PRESET_COLUMN_PATH, &old_path, -1);
	if (old_path)
		{
		new_path = radio_freq_to_path(radio_path_to_freq(old_path), newname);
		}
	else
		{
		new_path = radio_freq_to_path(preset_get_freq(tree_path_to_row(ted->path)), newname);
		}

	gtk_list_store_set(GTK_LIST_STORE(store), &iter,
			   PRESET_COLUMN_DESCRIPTION, newname,
			   PRESET_COLUMN_PATH, new_path, -1);

	g_free(old_path);
	g_free(new_path);

	return FALSE;
}

static gint preset_view_press_cb(GtkWidget *view, GdkEventButton *event, gpointer data)
{
	GtkTreeModel *store;
	GtkTreeViewColumn *tcolumn;
	GtkTreePath *tpath;
	GtkTreeIter iter;
	GList *list;
	gint row = -1;
	gint column;
	gint ret = FALSE;

	if (!gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(view), event->x, event->y,
					   &tpath, &tcolumn, NULL, NULL))
		{
		return FALSE;
		}

	row = tree_path_to_row(tpath);

	list = gtk_tree_view_get_columns(GTK_TREE_VIEW(view));
	column = g_list_index(list, tcolumn);
	g_list_free(list);

	store = gtk_tree_view_get_model(GTK_TREE_VIEW(view));
	gtk_tree_model_get_iter(store, &iter, tpath);

	if (column == PRESET_COLUMN_FREQUENCY)
		{
		gchar *text;
		gchar *old_path;

		gtk_tree_model_get(store, &iter,
				   PRESET_COLUMN_FREQUENCY, &text,
				   PRESET_COLUMN_PATH, &old_path, -1);

		if (event->button == 1)
			{
			tree_edit_by_path(GTK_TREE_VIEW(view), tpath, column, text,
					  preset_view_frequency_cb, NULL);

			ret = TRUE;
			}
		else if (event->button == 2 || event->button == 3 ||
			 event->button == 4 || event->button == 5)
			{
			guint32 freq;
			gchar *new_path;
			gchar *tmp;

			freq = (old_path) ? radio_path_to_freq(old_path) : preset_get_freq(row);

			if (event->button == 4 || event->button == 3)
				{
				freq += RADIO_SCAN_INCREMENT;
				}
			else
				{
				freq -= RADIO_SCAN_INCREMENT;
				}
			freq = radio_freq_clamp_to_increment(freq, RADIO_SCAN_INCREMENT);
			freq = CLAMP(freq, radio_limit_get_lower(), radio_limit_get_upper());

			tmp = (old_path) ? radio_path_to_title(old_path) : g_strdup(preset_get_description(row));
			new_path = radio_freq_to_path(freq, tmp);
			g_free(tmp);

			tmp = radio_freq_to_text(freq);
			gtk_list_store_set(GTK_LIST_STORE(store), &iter,
					   PRESET_COLUMN_FREQUENCY, tmp,
					   PRESET_COLUMN_PATH, new_path, -1);
			g_free(tmp);
			g_free(new_path);
			g_free(text);
			g_free(old_path);

			ret = TRUE;
			}
		}
	else if (column == 2)
		{
		if (event->button == 1)
			{
			gchar *text;

			gtk_tree_model_get(store, &iter, PRESET_COLUMN_DESCRIPTION, &text, -1);
			tree_edit_by_path(GTK_TREE_VIEW(view), tpath, column, text,
                                          preset_view_description_cb, NULL);
			g_free(text);

			ret = TRUE;
			}
		}

	gtk_tree_path_free(tpath);

	return ret;
}

static void preset_list_apply(GtkWidget *view)
{
	GtkTreeModel *store;
	GtkTreeIter iter;
	gint valid;
	gint i;
	gint change = FALSE;

	store = gtk_tree_view_get_model(GTK_TREE_VIEW(view));

	valid = gtk_tree_model_get_iter_first(store, &iter);
	i = 0;
	while (i < PRESET_LIST_SIZE && valid)
		{
		gchar *path;

		gtk_tree_model_get(store, &iter, PRESET_COLUMN_PATH, &path, -1);
		if (path)
			{
			if (strcmp(path, "clear") == 0)
				{
				preset_clear(i);
				}
			else
				{
				preset_set_by_path(i, path);
				}
			change = TRUE;
			g_free(path);
			}

		valid = gtk_tree_model_iter_next(store, &iter);
		i++;
		}

	if (change)
		{
		if (!preset_is_freq(preset, frequency))
			{
			preset = preset_find_freq(frequency);
			display_update_preset(FALSE, preset);
			}
		display_preset_list_update(-1);
		}
}

static void populate_preset_list(GtkListStore *store)
{
	gint i;

	for (i = 0; i < PRESET_LIST_SIZE; i++)
		{
		GtkTreeIter iter;
		guint32 freq;
		gchar *freq_text = NULL;
		const gchar *desc;

		freq = preset_get_freq(i);
		desc = preset_get_description(i);

		if (freq > 0)
			{
			freq_text = radio_freq_to_text(preset_get_freq(i));
			}

		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter,
				   PRESET_COLUMN_NUMBER, i + 1,
				   PRESET_COLUMN_FREQUENCY, freq_text,
				   PRESET_COLUMN_DESCRIPTION, desc,
				   PRESET_COLUMN_PATH, NULL, -1);


		g_free(freq_text);
		}
}

static void mixer_device_id_cb(GtkWidget *w, gpointer data)
{
	gchar *dev_text = data;
	gint d = get_mixer_device_id(dev_text);
	if (d >= 0) c_mixer_device_id = d;
}

static void spin_cb(GtkObject *adj, gpointer data)
{
	gint *c_n = data;

	*c_n = (gint)(GTK_ADJUSTMENT(adj)->value) * 10000;
}

static void skin_item_selected(GtkTreeSelection *selection, gpointer data)
{
	GtkTreeModel *store;
	GtkTreeIter iter;
	gchar *path;

	if (!gtk_tree_selection_get_selected(selection, &store, &iter)) return;

	gtk_tree_model_get(store, &iter, SKIN_COLUMN_PATH, &path, -1);
	gtk_entry_set_text(GTK_ENTRY(default_skin_entry), (path) ? path : "");
	g_free(path);
}

static gint sort_skin_list_cb(void *a, void *b)
{
	const gchar *fa = filename_from_path((gchar *)a);
	const gchar *fb = filename_from_path((gchar *)b);
	if (!fa || !fb) return 0;
	return strcmp(fa, fb);
}

static GList *skin_list_add_dir(GList *list, const gchar *skinpath)
{
	GList *dirs;

	if (!path_list(skinpath, NULL, &dirs)) return list;

	return g_list_concat(list, dirs);
}

static void populate_skin_list(GtkListStore *store)
{
	gchar *skinpath;
	GList *skin_list = NULL;
	GList *work;
	GtkTreeIter iter;

	/* add default skin */
	gtk_list_store_append(store, &iter);
	gtk_list_store_set(store, &iter,
			   SKIN_COLUMN_PATH, NULL,
			   SKIN_COLUMN_NAME, _("None (default)"), -1);

	/* add home dir skins */
	skinpath = g_strconcat(homedir(), "/", GQRADIO_RC_DIR_SKIN, NULL);
	skin_list = skin_list_add_dir(skin_list, skinpath);
	g_free(skinpath);

	/* add system dir skins */
	skin_list = skin_list_add_dir(skin_list, GQRADIO_SKINDIR);

	skin_list = g_list_sort(skin_list, (GCompareFunc) sort_skin_list_cb);

	work = skin_list;
	while (work)
		{
		gchar *path = work->data;
		gchar *skindata_file;

		work = work->next;

		if (strcmp(path, GQRADIO_SKINDIR "/default") == 0)
			{
			continue;
			}

		skindata_file = g_strconcat(path, "/skindata", NULL);
		if (isfile(skindata_file))
			{
			gtk_list_store_append(store, &iter);
			gtk_list_store_set(store, &iter,
					   SKIN_COLUMN_PATH, path,
					   SKIN_COLUMN_NAME, filename_from_path(path), -1);
			}
		g_free(skindata_file);
		}

	path_list_free(skin_list);
}

void config_window_show(gint start_tab)
{
	if (configwindow)
		{
		gdk_window_raise(configwindow->window);
		return;
		}

	create_config_window(start_tab);
	gtk_widget_show_all(configwindow);
}

static void config_window_hide(void)
{
	gtk_widget_destroy(configwindow);
	configwindow = NULL;
}

static void config_window_close_cb(GtkWidget *widget, gpointer data)
{
	config_window_hide();
}

static gint config_window_delete_cb(GtkWidget *widget, GdkEventAny *event, gpointer data)
{
	config_window_hide();
	return TRUE;
}

static void config_window_apply_cb(GtkWidget *widget, gpointer data)
{
	const gchar *tmp;
	gchar *buf = NULL;
	gint skin_changed = FALSE;
	gint need_init = FALSE;

	if (slik_double_size != c_slik_double_size) skin_changed = TRUE;

	if (slik_colorshift_on != c_slik_colorshift_on)
		{
		slik_colorshift_on = c_slik_colorshift_on;
		skin_changed = TRUE;
		}

	if (slik_colorshift_r != c_slik_colorshift_r ||
	    slik_colorshift_g != c_slik_colorshift_g ||
	    slik_colorshift_b != c_slik_colorshift_b ||
	    slik_colorshift_a != c_slik_colorshift_a )
		{
		slik_colorshift_r = c_slik_colorshift_r;
		slik_colorshift_g = c_slik_colorshift_g;
		slik_colorshift_b = c_slik_colorshift_b;
		slik_colorshift_a = c_slik_colorshift_a;
		if (slik_colorshift_on) skin_changed = TRUE;
		}

	if (slik_transparency_force != c_slik_transparency_force)
		{
		slik_transparency_force = c_slik_transparency_force;
		skin_changed = TRUE;
		}

	if (slik_transparency_force_a != c_slik_transparency_force_a)
		{
		slik_transparency_force_a = c_slik_transparency_force_a;
		if (slik_transparency_force) skin_changed = TRUE;
		}

	slik_double_size = c_slik_double_size;

	buf = remove_trailing_slash(gtk_entry_get_text(GTK_ENTRY(default_skin_entry)));
	if ( (buf && main_window->skin_path && strcmp(buf, main_window->skin_path) != 0 && strlen(buf) > 0) ||
		(!main_window->skin_path && buf && strlen(buf) > 0) )
		{
		/* change to the new skin */
		ui_skin_load(main_window, buf, NULL);
		tab_completion_append_to_history(default_skin_entry, buf);
		}
	else if (main_window->skin_path && buf && strlen(buf) == 0)
		{
		ui_skin_load(main_window, NULL, NULL);
		}
	else if (skin_changed)
		{
		ui_skin_load(main_window, main_window->skin_path, main_window->skin_mode_key);
		}
	g_free(buf);

	preset_list_apply(preset_view);

	tmp = gtk_entry_get_text(GTK_ENTRY(device_entry));
	if (( (strlen(tmp) > 0) != (radio_custom_device != NULL) ) ||
	    (radio_custom_device && strcmp(tmp, radio_custom_device) != 0) )
		{
		g_free(radio_custom_device);
		if (strlen(tmp) > 0)
			{
			radio_custom_device = g_strdup(tmp);
			}
		else
			{
			radio_custom_device = NULL;
			}
		need_init = TRUE;
		}

	freq_step = c_freq_step;

	if (mixer_device_id != c_mixer_device_id)
		set_mixer_device(c_mixer_device_id);
	mixer_device_id = c_mixer_device_id;

	if (mixer_command)
		{
		g_free(mixer_command);
		mixer_command = NULL;
		}
	tmp = gtk_entry_get_text(GTK_ENTRY(mixer_entry));
	if (tmp && strlen(tmp) > 0) mixer_command = g_strdup(tmp);

	slik_smart_placement = c_slik_smart_placement;
	slik_remember_position = c_slik_remember_position;
	wm_decorations = c_wm_decorations;

	if (description_scrolls_always != c_description_scrolls_always)
		{
		description_scrolls_always = c_description_scrolls_always;
		display_set_description(NULL);
		}

	skinned_menus_enable = c_skinned_menus_enable;

	gdk_window_set_decorations (main_window->window->window, wm_decorations);

	radio_limit_by_card_set(c_use_reported_freq_limits);

	if (need_init)
		{
		radio_shutdown();
		radio_set_device(radio_custom_device);
		if (radio_test(TRUE))
			{
			radio_startup(frequency);
			if (main_window->skin_mode_key && strcmp(main_window->skin_mode_key, "skindata_error") == 0)
				{	
				ui_skin_mode_set(main_window, NULL);
				}
			}
		else
			{
			buf = g_strdup_printf(_("Failed to start %s"), (radio_custom_device) ? radio_custom_device : RADIO_DEVICE);
			warning_dialog(_("Radio initialization error"), buf);
			g_free(buf);
			}
		}
}

static void config_window_ok_cb(GtkWidget *widget, gpointer data)
{
	config_window_apply_cb(widget, data);
	config_window_hide();
}

static void create_config_window(gint start_tab)
{
	GtkWidget *configvbox;
	GtkWidget *configtabs;
	GtkWidget *button;
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *vbox1;
	GtkWidget *label;
	GtkWidget *frame;
	GtkListStore *store;
	GtkWidget *skin_view;
	GtkWidget *sep;

        GtkWidget *omenu;
        GtkWidget *menu;
	GList *devices;
	GList *list;
	gint lh, cnt;

	GtkWidget *tabcomp;

	GtkObject *adj;
	GtkWidget *spin;
	gchar *text;

	configwindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_type_hint(GTK_WINDOW(configwindow), GDK_WINDOW_TYPE_HINT_DIALOG);
	gtk_window_set_wmclass(GTK_WINDOW(configwindow), "preferences", "GQradio");
	g_signal_connect(G_OBJECT(configwindow), "delete_event",
			 G_CALLBACK(config_window_delete_cb), NULL);
	gtk_window_set_resizable(GTK_WINDOW(configwindow), TRUE);
	gtk_window_set_default_size(GTK_WINDOW(configwindow), PREF_WINDOW_DEFAULT_WIDTH, PREF_WINDOW_DEFAULT_HEIGHT);
	gtk_window_set_title (GTK_WINDOW(configwindow), _("GQradio preferences"));
	gtk_container_set_border_width(GTK_CONTAINER(configwindow), 5);

	configvbox = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(configwindow),configvbox);
	gtk_widget_show(configvbox);

	hbox = gtk_hbutton_box_new();
	gtk_box_set_spacing(GTK_BOX(hbox), 10);
	gtk_box_pack_end(GTK_BOX(configvbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	button = gtk_button_new_from_stock(GTK_STOCK_OK);
	g_signal_connect(G_OBJECT (button), "clicked",
			 G_CALLBACK(config_window_ok_cb), NULL);
	gtk_container_add(GTK_CONTAINER(hbox), button);
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button);
	gtk_widget_show(button);

	button = gtk_button_new_from_stock(GTK_STOCK_APPLY);
	g_signal_connect(G_OBJECT(button), "clicked",
			 G_CALLBACK(config_window_apply_cb), NULL);
	gtk_container_add(GTK_CONTAINER(hbox), button);
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_show(button);

	button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	g_signal_connect(G_OBJECT(button), "clicked",
			 G_CALLBACK(config_window_close_cb), NULL);
	gtk_container_add(GTK_CONTAINER(hbox), button);
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_show(button);

	configtabs = gtk_notebook_new();
	gtk_notebook_set_tab_pos (GTK_NOTEBOOK(configtabs), GTK_POS_TOP);
	gtk_box_pack_start (GTK_BOX(configvbox), configtabs, TRUE, TRUE, 0);

	/* general tab */

	frame = gtk_frame_new(NULL);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
	gtk_widget_show(frame);
	label = gtk_label_new(_("General"));
	gtk_notebook_append_page (GTK_NOTEBOOK (configtabs), frame, label);

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	gtk_container_add (GTK_CONTAINER(frame),vbox);
	gtk_widget_show(vbox);
	
	frame = gtk_frame_new(_("Window"));
	gtk_container_set_border_width (GTK_CONTAINER (frame), 5);
	gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);

	vbox1 = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER(frame),vbox1);
	gtk_widget_show(vbox1);

	button = checkbox_new(_("Smart window placement"), slik_smart_placement, &c_slik_smart_placement);
	gtk_box_pack_start(GTK_BOX(vbox1), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	button = checkbox_new(_("Remember window position"), slik_remember_position, &c_slik_remember_position);
	gtk_box_pack_start(GTK_BOX(vbox1), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	button = checkbox_new(_("Enable window manager decorations"), wm_decorations, &c_wm_decorations);
	gtk_box_pack_start(GTK_BOX(vbox1), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	button = checkbox_new(_("Double sized window"), slik_double_size, &c_slik_double_size);
	gtk_box_pack_start(GTK_BOX(vbox1), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	button = checkbox_new(_("Always scroll description"), description_scrolls_always, &c_description_scrolls_always);
	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	frame = gtk_frame_new(_("Mixer"));
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
	gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);

	vbox1 = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER(frame),vbox1);
	gtk_widget_show(vbox1);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	label = gtk_label_new(_("External Mixer:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show (label);

	gtk_widget_realize(configwindow);

	tabcomp = tab_completion_new(&mixer_entry, mixer_command, NULL, NULL);
	gtk_box_pack_start(GTK_BOX(hbox), tabcomp,TRUE,TRUE,5);
	gtk_widget_show(tabcomp);

	c_mixer_device_id = mixer_device_id;
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	label = gtk_label_new(_("Volume controls device:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show (label);

        omenu = gtk_option_menu_new ();
        menu = gtk_menu_new ();

	devices = get_mixer_device_list();

	list = devices;

	lh = 0;
	cnt = 0;
	while (list)
		{
		gchar *dev_text = list->data;

		menu_item_add(menu, dev_text,
			      G_CALLBACK(mixer_device_id_cb), dev_text);

		if (get_mixer_device_id(dev_text) == mixer_device_id) lh = cnt;
		cnt++;
		list = list->next;
		}

	g_list_free(devices);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (omenu), menu);
        gtk_option_menu_set_history (GTK_OPTION_MENU (omenu), lh);
        gtk_box_pack_start( GTK_BOX(hbox), omenu, TRUE, TRUE, 5);
        gtk_widget_show (omenu);

	/* tuner tab */

	frame = gtk_frame_new(NULL);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
	gtk_widget_show(frame);
	label = gtk_label_new(_("Tuner"));
	gtk_notebook_append_page (GTK_NOTEBOOK (configtabs), frame, label);

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	gtk_container_add (GTK_CONTAINER(frame),vbox);
	gtk_widget_show(vbox);
	
	frame = gtk_frame_new(_("Device"));
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
	gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);

	vbox1 = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER(frame),vbox1);
	gtk_widget_show(vbox1);

	device_entry = gtk_entry_new();
	if (radio_custom_device) gtk_entry_set_text(GTK_ENTRY(device_entry), radio_custom_device);
	gtk_box_pack_start(GTK_BOX(vbox1), device_entry, FALSE, FALSE, 0);
	gtk_widget_show(device_entry);

	button = checkbox_new(_("Use frequency limits reported by driver"), radio_limit_by_card_get(), &c_use_reported_freq_limits);
	gtk_box_pack_start(GTK_BOX(vbox1), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	hbox = gtk_hbox_new (FALSE, 5);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	label = gtk_label_new(_("Frequency step (x10Khz):"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	c_freq_step = freq_step;
	adj = gtk_adjustment_new((gfloat)freq_step / 10000.0, 1.0, 100.0, 1.0, 1.0, 1.0);
	spin = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 1.0, 0);
	g_signal_connect(adj, "value_changed",
			   G_CALLBACK(spin_cb), &c_freq_step);
	gtk_box_pack_start(GTK_BOX(hbox), spin, FALSE, FALSE, 0);
	gtk_widget_show(spin);

	/* preset tab */

	frame = gtk_frame_new(NULL);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
	gtk_widget_show(frame);
	label = gtk_label_new(_("Presets"));
	gtk_notebook_append_page (GTK_NOTEBOOK (configtabs), frame, label);

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	gtk_container_add (GTK_CONTAINER(frame),vbox);
	gtk_widget_show(vbox);

	store = gtk_list_store_new(4, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	preset_view = ui_edit_list_new(vbox, store, NULL, NULL);
	g_object_unref(store);

	ui_edit_list_add_column(preset_view, _("#"), PRESET_COLUMN_NUMBER, FALSE, FALSE);
	ui_edit_list_add_column(preset_view, _("Frequency"), PRESET_COLUMN_FREQUENCY, FALSE, FALSE);
	ui_edit_list_add_column(preset_view, _("Description"), PRESET_COLUMN_DESCRIPTION, FALSE, FALSE);

	g_signal_connect(G_OBJECT(preset_view), "button_press_event",
			 G_CALLBACK(preset_view_press_cb), NULL);

	populate_preset_list(store);

	/* skin tab */

	frame = gtk_frame_new(NULL);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
	gtk_widget_show(frame);
	label = gtk_label_new(_("Skins"));
	gtk_notebook_append_page (GTK_NOTEBOOK (configtabs), frame, label);

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	gtk_container_add (GTK_CONTAINER(frame),vbox);
	gtk_widget_show(vbox);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	label = gtk_label_new(_("Skin:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
	gtk_widget_show (label);

	tabcomp = tab_completion_new_with_history(&default_skin_entry,
						  main_window->skin_path ? main_window->skin_path : "",
						  "select_skin", 16, NULL, NULL);
	gtk_box_pack_start(GTK_BOX(hbox), tabcomp, TRUE, TRUE, 0);
	gtk_widget_show(tabcomp);

	/* skin list */
	store = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_STRING);
	skin_view = ui_edit_list_new(vbox, store,
				     G_CALLBACK(skin_item_selected), NULL);
	g_object_unref(store);

	ui_edit_list_add_column(skin_view, _("Available skins:"), SKIN_COLUMN_NAME, FALSE, FALSE);

	populate_skin_list(store);

	text = g_strconcat(homedir(), "/", GQRADIO_RC_DIR_SKIN, NULL);
	label = gtk_label_new(text);
	g_free(text);

	gtk_box_pack_end(GTK_BOX(vbox), label, FALSE, FALSE, 0);
	gtk_widget_show (label);

	frame = gtk_frame_new(_("Skin Adjustment"));
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
	gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);

	vbox1 = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER(frame), vbox1);
	gtk_widget_show(vbox1);

	hbox = ui_edit_frame_new(vbox1, FALSE, NULL);

	button = checkbox_new(_("Color tint:"), slik_colorshift_on, &c_slik_colorshift_on);
	gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	hscale_new(hbox, NULL, slik_colorshift_a, &c_slik_colorshift_a, FALSE);

	hscale_new(vbox1, "R", slik_colorshift_r, &c_slik_colorshift_r, TRUE);
	hscale_new(vbox1, "G", slik_colorshift_g, &c_slik_colorshift_g, TRUE);
	hscale_new(vbox1, "B", slik_colorshift_b, &c_slik_colorshift_b, TRUE);

	hbox = ui_edit_frame_new(vbox1, FALSE, NULL);

	button = checkbox_new(_("Transparent:"), slik_transparency_force, &c_slik_transparency_force);
	gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	hscale_new(hbox, NULL, slik_transparency_force_a, &c_slik_transparency_force_a, FALSE);

	button = checkbox_new(_("Enable menu skins"), skinned_menus_enable, &c_skinned_menus_enable);
	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	sep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 0);
	gtk_widget_show(sep);

	gtk_notebook_set_current_page(GTK_NOTEBOOK(configtabs), start_tab);

	gtk_widget_show(configtabs);
}

